private import common;
private import base64;

private import SDL_net;

private import std.stream;
private import std.string;
private import std.ctype;
private import std.uri;
private import std.conv;
private import std.c.stdio;

class Proxy {
    static char[] host;
    static int port;
    static char[] user;
    static char[] pass;
}

class HttpError : Error {
    this(char[] msg) {
        super(msg ~ ": " ~ tostr(SDLNet_GetError()));
    }
}

class HttpResponse {
public:
    this(TCPsocket sock) {
        char[] data;
        char[8192] buf;
        int len = SDLNet_TCP_Recv(sock, &buf[0], buf.length);
        while (len != 0) {
            data ~= buf[0..len];
            len = SDLNet_TCP_Recv(sock, &buf[0], buf.length);
        }
        data ~= buf[0..len];

        int bodyStart = data.find("\r\n\r\n");
        // it's HTTP/0.9
        if (bodyStart == -1) {
            messageBody = data;
            return;
        }

        messageBody = data[bodyStart+4..length];

        char[][] lines = split(data[0..bodyStart], "\r\n");

        statusLine = lines[0];

        int index = 0;
        while (index < statusLine.length && !isspace(statusLine[index]))
            index++;

        // it's HTTP/0.9
        if (index < 4 || statusLine[0..4] != "HTTP") {
            messageBody = data;
            return;
        }

        httpVersion = statusLine[0..index];

        while (isspace(statusLine[index])) index++;
        int index2 = index;

        while (!isspace(statusLine[index])) index++;
        code = statusLine[index2..index];

        while (isspace(statusLine[index])) index++;
        message = statusLine[index..statusLine.length];

        len = -1;
        int i = 1;
        bool transfer_encoding_chunked = false;
        char[] line = lines[i];
        while (line != "") {
            int f = find(line, ": ");
            char[] k = line[0..f];
            char[] v = line[f+2..line.length];

            char[] lk = std.string.tolower(k);
            if (lk == `content-length`) {
                len = toInt(v);
            }
            else if (lk == `transfer-encoding`) {
                if (std.string.tolower(v) == `chunked`) {
                    transfer_encoding_chunked = true;
                }
            }

            header[k] = v;

            i++;
            if (i == lines.length) break;
            line = lines[i];
        }

        if (len != -1 && len != messageBody.length) {
            printf("%d %d\n", len, messageBody.length);
            throw new HttpError("length missmatch");
        }
    }

	char[] toString() {
        if (all) return all;

        all = statusLine ~ "\r\n";

        char[][] keys = header.keys;

        foreach (char[] k; header.keys) {
            all ~= k ~ ": " ~ header[k] ~ "\r\n";
        }

        all ~= "\r\n" ~ messageBody;

        return all;
	}

public:
    char[] all;
	char[] statusLine;
    char[] httpVersion;
    char[] code;
    char[] message;
	char[][char[]] header;
	char[] messageBody;
}

class Http {
    this(char[] host, int port) {
        host_ = host;
        port_ = port;
    }

    abstract HttpResponse get_(char[] uri);

    HttpResponse get(char[] uri) { return get_(fullUri(uri)); }
    HttpResponse get(char[] uri, char[] q) {
        return get(fullUri(uri) ~ "?" ~ q);
    }
    HttpResponse get(char[] uri, char[][char[]] q) {
        return get(fullUri(uri), joinQuery(q));
    }

    abstract HttpResponse post_(char[] uri, char[] q);

    HttpResponse post(char[] uri, char[] q) { return post_(fullUri(uri), q); }
    HttpResponse post(char[] uri, char[][char[]] q) {
        return post(fullUri(uri), joinQuery(q));
    }

protected:
    char[] fullUri(char[] uri) {
        if (uri.length < 7 || uri[0..7] != "http://") {
            if (port_ == 80) {
                return "http://" ~ host_ ~ uri;
            }
            else {
                return "http://" ~ host_ ~ ":" ~ tostr(port_) ~ uri;
            }
        }
        return uri;
    }

    char[] joinQuery(char[][char[]] q) {
        char[] qs;
        foreach (char[] k, char[] v; q) {
            qs ~= k ~ "=" ~ v ~ "&";
        }
        qs.length = qs.length - 1;
        return std.uri.encode(qs);
    }

    char[] createCommonHeader() {
        char[] headerStr = ``;

        if (Proxy.user != ``) {
            char[512] buf;
            char[] up = Proxy.user~":"~Proxy.pass;
            mutt_to_base64(cast(ubyte*)&buf[0], cast(ubyte*)&up[0],
                           up.length, 512);
            headerStr ~=
                "Proxy-Authorization: Basic " ~ tostr(cast(char*)buf) ~ "\r\n";
        }

        headerStr ~= "Host: " ~ host_ ~ "\r\n";
        foreach (char[] k, char[] v; header) {
            headerStr ~= k ~ ": " ~ v ~ "\r\n";
        }

        return headerStr;
    }

    TCPsocket connect() {
        IPaddress ip;
        if (Proxy.host != ``) {
            SDLNet_ResolveHost(&ip, toStringz(Proxy.host), Proxy.port);
        }
        else {
            SDLNet_ResolveHost(&ip, toStringz(host_), port_);
        }

        if (ip.host == INADDR_NONE) {
            throw new HttpError(`resolve failed`);
        }

        TCPsocket sock = SDLNet_TCP_Open(&ip);

        if (sock == null) {
            throw new HttpError(`connect failed`);
        }

        return sock;
    }

    HttpResponse request(char[] req) {
        TCPsocket ss = connect();

        SDLNet_TCP_Send(ss, toStringz(req), req.length);
//        Log.info(req);

        HttpResponse res = new HttpResponse(ss);

        SDLNet_TCP_Close(ss);

        return res;
    }

protected:
    char[] host_;
    int port_;
    char[][char[]] header;
}

class Http09 : Http {
    this(char[] host, int port) {
        super(host, port);
    }

    override HttpResponse get_(char[] uri) {
        char[] req = "GET " ~ uri ~ "\r\n";

        return request(req);
    }

    override HttpResponse post_(char[] uri, char[] q) {
        // not implemented
        assert(false);
        return null;
    }
}

class Http10 : Http {
    this(char[] host, int port) {
        super(host, port);
    }

    override HttpResponse get_(char[] uri) {
        char[] req =
            "GET " ~ uri ~ " HTTP/1.0\r\n" ~
            createCommonHeader() ~ "\r\n";

        return request(req);
    }

    override HttpResponse post_(char[] uri, char[] q) {
        char[] req =
            "POST " ~ uri ~ " HTTP/1.0\r\n" ~
            createCommonHeader() ~
            "Content-Length: " ~ std.string.toString(q.length) ~ "\r\n"
            "\r\n" ~ q;

        return request(req);
    }
}

class Http11 : Http {
    this(char[] host, int port) {
        super(host, port);
        close_ = false;
    }

    override HttpResponse get_(char[] uri) {
        char[] req =
            "GET " ~ uri ~ " HTTP/1.1\r\n" ~
            createCommonHeader() ~
            "Connection: " ~ ((close_) ? "close" : "keep-alive") ~ "\r\n"
            "\r\n";

        if (!ss_) ss_ = connect();

        SDLNet_TCP_Send(ss_, toStringz(req), req.length);

        HttpResponse res = new HttpResponse(ss_);

        if (std.string.tolower(res.header[`Connection`]) != `keep-alive`) {
            SDLNet_TCP_Close(ss_);
            ss_ = null;
        }

        return res;
    }

    override HttpResponse post_(char[] uri, char[] q) {
        char[] req =
            "POST " ~ uri ~ " HTTP/1.0\r\n" ~
            createCommonHeader() ~
            "Connection: " ~ ((close_) ? "close" : "keep-alive") ~ "\r\n"
            "Content-Length: " ~ std.string.toString(q.length) ~ "\r\n"
            "\r\n" ~ q;

        if (!ss_) ss_ = connect();

        SDLNet_TCP_Send(ss_, toStringz(req), req.length);

        HttpResponse res = new HttpResponse(ss_);

        if (std.string.tolower(res.header[`Connection`]) != `keep-alive`) {
            SDLNet_TCP_Close(ss_);
            ss_ = null;
        }

        return res;
    }

    bool alive() { return ss_ !== null; }

    bool close() { return close; }
    void close(bool c) { close_ = c; }

private:
    TCPsocket ss_;
    bool close_;
}

version (USE_UNIT) {
unittest {
    Http http;

    http = new Http09("localhost", 80);
    std.stream.stdout.writeLine(http.get("/~i/env.cgi", "hogehoge").toString);

    http = new Http10("localhost", 80);
    std.stream.stdout.writeLine(http.get("/~i/env.cgi", "hogehoge").toString);

    http = new Http10("localhost", 80);
    std.stream.stdout.writeLine(http.post("/~i/env.cgi", "hogehoge").toString);

    Http11 http11;
    http11 = new Http11("localhost", 80);
    std.stream.stdout.writeLine(http11.get("/~i/env.cgi").toString);
    assert(http11.alive);
    http11.close = true;
    std.stream.stdout.writeLine(http11.get("/~i/env.cgi").toString);
    assert(!http11.alive);
}
}

